from typing import Callable, Dict, Optional, Tuple, Union

from transformers import PreTrainedModel, PreTrainedTokenizer

from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import ButtonElement, ChoicesSubElement

TOKENIZER_MODEL_TUPLE = Tuple[PreTrainedTokenizer, PreTrainedModel]
MODEL_TOKENIZER_CHOICES = Union[
    Dict[str, TOKENIZER_MODEL_TUPLE],
    Dict[
        str,
        Callable[[], TOKENIZER_MODEL_TUPLE],
    ],
]


class ModelSelectionMixin:
    def __init__(
        self,
        on_model_change_callback: Callable[[], None],
        model: Optional[PreTrainedModel] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        model_tokenizer_choices: Optional[MODEL_TOKENIZER_CHOICES] = None,
        keep_models_in_memory: bool = True,
    ):
        if model is None and (
            model_tokenizer_choices is None or len(model_tokenizer_choices) == 0
        ):
            raise ValueError()

        if model is None or tokenizer is None:
            assert model_tokenizer_choices is not None
            self._model_choices = model_tokenizer_choices
        else:
            assert model_tokenizer_choices is None
            self._model_choices = None
            self._model, self._tokenizer = model, tokenizer

        self._cache: Optional[Dict[str, TOKENIZER_MODEL_TUPLE]] = None
        if keep_models_in_memory:
            self._cache = {}
        self._on_model_change_callback = on_model_change_callback
        self.initialize_model_elements()

        if model_tokenizer_choices is not None:
            self._tokenizer, self._model = self.load_model(
                model_tokenizer_choices[self.model_selector_element.selected]
            )

    def load_model(
        self,
        value: Union[
            TOKENIZER_MODEL_TUPLE,
            Callable[[], TOKENIZER_MODEL_TUPLE],
        ],
    ):
        if isinstance(value, tuple):
            return value
        else:
            return value()

    def initialize_model_elements(self):
        if self._model_choices is None:
            return

        self.model_selector_heading = PlainTextElement(
            content="Model Settings", is_heading=True
        )
        self.model_selector_element = ChoicesSubElement(
            list(self._model_choices), text="Select Model"
        )
        self.button_element = ButtonElement(
            processing_callback=self.model_callback,
            button_text="Send Model Configuration",
            subelements=[self.model_selector_element],
        )

    @property
    def model_elements(self):
        if self._model_choices is None:
            return []
        return [self.model_selector_heading, self.button_element]

    def model_callback(self):
        assert self._model is not None
        if self._model_choices is None:
            return
        if self.model_selector_element.updated:
            key = self.model_selector_element.selected
            if self._cache is not None and key in self._cache:
                self._tokenizer, self._model = self._cache[key]
            else:
                self._tokenizer, self._model = self.load_model(self._model_choices[key])
                if self._cache is not None:
                    self._cache[key] = (self._tokenizer, self._model)
            self._on_model_change_callback()



from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union

from flask import request

from visuallm.server import Server

from .element_base import ElementDescription, ElementWithEndpoint


@dataclass
class BarInfo:
    barTitle: str
    barHeights: List[float]
    barAnnotations: List[str]

    def __post_init__(self):
        for height in self.barHeights:
            if (height < 0) or (height > 100):
                raise ValueError(
                    f"Height should be between 0 and 100 (currently {height})"
                )


class BarChartElement(ElementWithEndpoint):
    def __init__(
        self,
        long_contexts: bool = False,
        names: List[str] = [],
        processing_callback: Optional[Callable[[], Any]] = None,
        name="barchart",
        **kwargs,
    ):
        """
        Args:
            `long_contexts` (bool, optional): If this flag is set, then the
                bar is displayed under the bar title, if it is not set, then the
                bar is displayed alongside the bar title. Defaults to False.
            `names` (List[str], optional): Names of the individual bars. These
                names are displayed above each individual bar if there are
                more than 1 bar or if `long_contexts` is set. There should be
                equal number of names to bar_heights. Defaults to [].
            `processing_callback` (Optional[Callable[[], None]): the callback
                function will be called just after the `BarChartElement`
                updated `self.selected,` so you can handle the change.
                Defaults to None, if `processing_callback == None` then no
                `Select` button will display on the frontend.
            `name` (str, optional): name of the element, doesn't have to be
                provided. Defaults to "barchart".
        """

        super().__init__(name=name, endpoint_callback=self.default_callback, **kwargs)

        self.processing_callback = processing_callback
        self._possibilities: List[BarInfo] = []
        self._selected: Optional[str] = None
        self.long_contexts = long_contexts
        self.names = names

    @property
    def possibilities(self) -> List[BarInfo]:
        return self._possibilities

    @possibilities.setter
    def possibilities(self, value: List[BarInfo]):
        self.changed = True
        self._possibilities = value

    @property
    def selected(self) -> str:
        assert self._selected is not None
        return self._selected

    @property
    def selectable(self):
        return self.processing_callback is not None

    def set_possibilities(
        self,
        bar_heights: Union[List[List[float]], Any],
        bar_annotations: List[List[str]],
        annotations: List[str],
    ):
        """
        Args:
            bar_heights (Union[List[List[float]], Any]): List of bar heights for each bar to be displayed.
                e.g. if the shape of list is len(bar_heigths) = N, and len(bar_heights[i]) = M for each i,
                then N lines each with M bars will be displayed.
            bar_annotations (List[List[str]]): List of bar annotations for each bar to be displayed.
                e.g. as in the previous example, bar_annotations[i][j] would be the text in
                j-th bar in i-th line
            annotations (List[str]): List of annotations for each group of bars to be displayed on
                single line.
        """
        self.possibilities = [
            BarInfo(annot, bar_height, bar_annot)
            for annot, bar_height, bar_annot in zip(
                annotations, bar_heights, bar_annotations
            )
        ]

    def check_possibilities_length(self):
        required_len = len(self.names)
        if required_len == 0:
            # if there isn't any name provided, then populate self.names
            # with number of empty strings equal to size of first barHeights
            if len(self.possibilities) != 0:
                required_len = len(self.possibilities[0].barHeights)
                self.names = ["" for _ in range(required_len)]
            else:
                self.names = []

        for p in self.possibilities:
            for arr, name in [
                (p.barHeights, "Bar Heights"),
                (p.barAnnotations, "Bar Annotations"),
            ]:
                if len(arr) != required_len:
                    raise ValueError(
                        f"{name}: {arr} ({len(arr)}), names: {self.names}"
                        + f" ({len(self.names)})"
                    )

    def construct_element_description(self):
        self.changed = False
        self.check_possibilities_length()
        return ElementDescription(
            name=self.name,
            type="softmax",
            configuration=dict(
                bar_infos=self.possibilities,
                address=self.endpoint_url.removeprefix("/"),
                long_contexts=self.long_contexts,
                names=self.names,
                selectable=self.selectable,
            ),
        )

    def add_endpoint(self, app: Server):
        app.add_endpoint(self.endpoint_url, self.endpoint_callback, methods=["POST"])

    def default_callback(self):
        if not request.is_json:
            raise RuntimeError()

        # just for the type checker
        if request.json is None:
            raise RuntimeError()

        self._selected = request.json.get("selected")
        assert self.processing_callback is not None
        self.processing_callback()

        assert self.parent_component is not None
        return self.parent_component.fetch_info(fetch_all=False)



from .element_base import ElementBase, ElementDescription


class PlainTextElement(ElementBase):
    def __init__(
        self,
        content: str = "",
        is_heading: bool = False,
        heading_level=3,
        name="plain_text",
    ):
        super().__init__(name=name)
        self._content = content
        self.is_heading = is_heading
        self.heading_level = heading_level

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        self.changed = True
        self._content = value

    def construct_element_description(self):
        self.changed = False
        return ElementDescription(
            name=self.name,
            type="plain",
            configuration=dict(
                value=self.content,
                heading=self.is_heading,
                heading_level=self.heading_level,
            ),
        )

    def add_endpoint(self, app):
        pass



from typing import Optional

from visuallm.component_base import ComponentBase
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.elements.plain_text_element import PlainTextElement


class DatasetVisualization(ComponentBase, DataPreparationMixin):
    def __init__(
        self,
        title: str = "Dataset Visualization",
        dataset: Optional[DATASET_TYPE] = None,
        dataset_choices: Optional[DATASETS_TYPE] = None,
    ):
        self.main_heading_element = PlainTextElement(
            is_heading=True, heading_level=2, content=title
        )
        sample_vis_elements = self.initialize_sample_vis_elements()
        DataPreparationMixin.__init__(
            self,
            on_sample_change_callback=self.on_sample_change_callback,
            dataset=dataset,
            dataset_choices=dataset_choices,
        )

        super().__init__(
            name="dataset_visualization",
            title=title,
            elements=[
                self.main_heading_element,
                *self.dataset_elements,
                *sample_vis_elements,
            ],
        )

    def initialize_sample_vis_elements(self):
        self.sample_vis_element = PlainTextElement()
        return [self.sample_vis_element]

    def update_sample_vis_elements(self):
        self.sample_vis_element.content = self._loaded_sample

    def on_sample_change_callback(self):
        self.update_sample_vis_elements()



from typing import Any, Callable, Dict, List, Sequence, Tuple

from visuallm.elements.barchart_element import BarChartElement
from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import ButtonElement, CheckBoxSubElement

GENERATED_TEXT_METRIC = Tuple[Callable[[str, str], Any], str, bool]
"""Args:
Callable: Metric which will be called to evaluate
str: Format of the value that should be displayed on the page, e.g.
    "{:.4f}" so that only 4 floating point numbers would be displayed
bool: Whether the bar is also scalable, or should be kept at 100 for the
    metric number to be better visible
"""

PROBS_METRIC = Tuple[Callable[[Any, Any], Any], str, bool]
"""Args:
Callable: Metric which will be called to evaluate
str: Format of the value that should be displayed on the page, e.g.
    "{:.4f}" so that only 4 floating point numbers would be displayed
bool: Whether the bar is also scalable, or should be kept at 100 for the
    metric number to be better visible
"""


class MetricsMixin:
    def __init__(
        self,
        on_metrics_change: Callable[[], None],
        metrics_on_generated_text: Dict[str, GENERATED_TEXT_METRIC] = {},
        metrics_on_probs: Dict[str, PROBS_METRIC] = {},
    ):
        self._select_metrics_heading = PlainTextElement(
            content="Which Metrics to Display", is_heading=True
        )
        self._select_metrics_elements: Dict[str, CheckBoxSubElement] = {}

        for dct in [metrics_on_generated_text, metrics_on_probs]:
            for key in dct:
                self._select_metrics_elements[key] = CheckBoxSubElement(key, True)
        self._ordering = list(metrics_on_generated_text.keys()) + list(
            metrics_on_probs.keys()
        )
        self._metrics_on_generated_text = metrics_on_generated_text
        self._metrics_on_probs = metrics_on_probs
        self._display_metrics_heading = PlainTextElement(
            content="Metrics on Generated Outputs", is_heading=True
        )
        self._display_metrics_element = BarChartElement(
            long_contexts=True, names=self._ordering
        )
        self._display_metrics_on_target_heading = PlainTextElement(
            content="Metrics on Target", is_heading=True
        )
        self._display_metrics_on_target_element = BarChartElement(
            long_contexts=True, names=self._ordering
        )

        self.metric_button_element = ButtonElement(
            subelements=list(self._select_metrics_elements.values()),
            button_text="Select Metrics to Display",
            processing_callback=self.metrics_processing_callback,
        )
        self._on_metrics_change = on_metrics_change

    @property
    def metrics_selection(self):
        return [self._select_metrics_heading, self.metric_button_element]

    @property
    def metrics_display(self):
        return [
            self._display_metrics_on_target_heading,
            self._display_metrics_on_target_element,
            self._display_metrics_heading,
            self._display_metrics_element,
        ]

    def _compute_metrics_for_element(
        self,
        generated_text_list: Sequence[str],
        label_text_list: Sequence[str],
        probs_encoded_list: Sequence[Any],
        generated_encoded_list: Sequence[Any],
        element: BarChartElement,
    ):
        bar_annotations: List[List[str]] = []
        bar_heights: List[List[float]] = []
        annotations: List[str] = []
        names: List[str] = []
        first_time: bool = True
        for (
            generated_text,
            label_text,
            labels_encoded,
            generated_encoded,
        ) in zip(
            generated_text_list,
            label_text_list,
            probs_encoded_list,
            generated_encoded_list,
        ):
            group_annotations: List[str] = []
            group_heights: List[float] = []
            for name in self._ordering:
                if not self._select_metrics_elements[name].selected:
                    continue
                if first_time:
                    names.append(name)

                if name in self._metrics_on_generated_text:
                    (
                        metric,
                        format_of_annotation,
                        scale_bar,
                    ) = self._metrics_on_generated_text[name]
                    result = metric(generated_text, label_text)
                else:
                    (
                        metric,
                        format_of_annotation,
                        scale_bar,
                    ) = self._metrics_on_probs[name]
                    result = metric(labels_encoded, generated_encoded)

                if scale_bar:
                    group_heights.append(min(result * 100, 100))
                else:
                    group_heights.append(100)
                group_annotations.append(format_of_annotation.format(result))

            first_time = False
            bar_annotations.append(group_annotations)
            bar_heights.append(group_heights)
            annotations.append(generated_text)

        element.set_possibilities(bar_heights, bar_annotations, annotations)
        element.names = names

    def compute_metrics_on_predicted(
        self,
        generated_text_list: Sequence[str],
        label_text_list: Sequence[str],
        probs_encoded_list: Sequence[Any],
        generated_encoded_list: Sequence[Any],
    ):
        self._compute_metrics_for_element(
            generated_text_list,
            label_text_list,
            probs_encoded_list,
            generated_encoded_list,
            self._display_metrics_element,
        )

    def compute_metrics_on_target(
        self,
        target: str,
        probs_target: Any,
        target_encoded: Any,
    ):
        self._compute_metrics_for_element(
            [target],
            [target],
            probs_target,
            target_encoded,
            self._display_metrics_on_target_element,
        )

    def metrics_processing_callback(self):
        self._on_metrics_change()



from heapq import nlargest
from typing import Optional

import torch
from numpy.typing import NDArray
from transformers import PreTrainedModel, PreTrainedTokenizer

from visuallm.component_base import ComponentBase
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.components.mixins.model_selection_mixin import (
    MODEL_TOKENIZER_CHOICES,
    ModelSelectionMixin,
)
from visuallm.elements.barchart_element import BarChartElement
from visuallm.elements.plain_text_element import PlainTextElement


class NextTokenPredictionComponent(
    ComponentBase, ModelSelectionMixin, DataPreparationMixin
):
    def __init__(
        self,
        title: str = "Next Token Prediction",
        model: Optional[PreTrainedModel] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        model_tokenizer_choices: Optional[MODEL_TOKENIZER_CHOICES] = None,
        dataset: Optional[DATASET_TYPE] = None,
        dataset_choices: Optional[DATASETS_TYPE] = None,
        n_largest_tokens_to_return: int = 10,
    ):
        self.main_heading_element = PlainTextElement(
            is_heading=True,
            heading_level=2,
            content=title,
        )
        ModelSelectionMixin.__init__(
            self,
            model_tokenizer_choices=model_tokenizer_choices,
            model=model,
            tokenizer=tokenizer,
            on_model_change_callback=self.on_model_change_callback,
        )
        DataPreparationMixin.__init__(
            self,
            on_sample_change_callback=self.on_sample_change_callback,
            dataset=dataset,
            dataset_choices=dataset_choices,
        )
        softmax_elements = self.init_softmax_heading_elements()
        input_display_elements = self.init_model_input_display()
        self._n_largest_tokens_to_return = n_largest_tokens_to_return
        self.init_word_vocab()

        super().__init__(
            name="next_token_prediction",
            title=title,
            elements=[
                self.main_heading_element,
                *self.dataset_elements,
                *self.model_elements,
                *input_display_elements,
                *softmax_elements,
            ],
        )

    def init_word_vocab(self):
        vocab_size = len(self._tokenizer.get_vocab())
        self.word_vocab = [""] * vocab_size
        for str_val, int_val in self._tokenizer.get_vocab().items():
            self.word_vocab[int_val] = str_val

    def init_softmax_heading_elements(self):
        self.softmax_heading_element = PlainTextElement(
            content="Next Token Probabilities:", is_heading=True
        )
        self.softmax_element = BarChartElement(
            processing_callback=self.select_next_token,
        )
        return [self.softmax_heading_element, self.softmax_element]

    def init_model_input_display(self):
        self.input_display = PlainTextElement()
        return [self.input_display]

    def update_model_input_display_on_sample_change(self):
        self.input_display.content = self._loaded_sample

    def update_model_input_display_on_selected_token(self, detokenized_token: str):
        self.input_display.content += detokenized_token

    def create_model_inputs(self):
        model_inputs = self._tokenizer(self.input_display.content, return_tensors="pt")
        return model_inputs

    def on_model_change_callback(self):
        self.init_word_vocab()
        self.data_force_update()
        self.dataset_callback()

    def _one_token_prediction(self):
        model_inputs = self.create_model_inputs()
        with torch.no_grad():
            probs: torch.Tensor = self._model(**model_inputs).logits
            probs = torch.softmax(probs, dim=-1)
        np_probs: NDArray = probs[0, -1, :].numpy()
        return np_probs

    def _get_n_largest_tokens_and_probs(self, probs: NDArray):
        return nlargest(
            self._n_largest_tokens_to_return,
            zip(map(lambda x: float(x) * 100, probs), self.word_vocab),
            key=lambda x: x[0],
        )

    def update_softmax_element(self):
        probs = self._one_token_prediction()
        n_largest_probs = self._get_n_largest_tokens_and_probs(probs)

        bar_heights = [[x[0]] for x in n_largest_probs]
        bar_annotations = [[f"{x[0]: .3f}%"] for x in n_largest_probs]
        annotations = [x[1] for x in n_largest_probs]
        self.softmax_element.set_possibilities(
            bar_heights, bar_annotations, annotations
        )

    def select_next_token(self):
        token = self.softmax_element.selected
        detokenized_token = self._tokenizer.convert_tokens_to_string([token])
        self.update_model_input_display_on_selected_token(detokenized_token)
        self.update_softmax_element()

    def on_sample_change_callback(self):
        self.update_model_input_display_on_sample_change()
        self.update_softmax_element()



from typing import Dict, List, Optional, cast

import torch
from transformers import PreTrainedModel, PreTrainedTokenizer
from transformers.generation.utils import GenerateOutput
from transformers.tokenization_utils_base import BatchEncoding

from visuallm.component_base import ComponentBase
from visuallm.components.mixins.data_preparation_mixin import (
    DATASET_TYPE,
    DATASETS_TYPE,
    DataPreparationMixin,
)
from visuallm.components.mixins.generation_selectors_mixin import (
    SELECTORS_TYPE,
    GenerationSelectorsMixin,
)
from visuallm.components.mixins.metrics_mixin import (
    GENERATED_TEXT_METRIC,
    PROBS_METRIC,
    MetricsMixin,
)
from visuallm.components.mixins.model_selection_mixin import (
    MODEL_TOKENIZER_CHOICES,
    ModelSelectionMixin,
)
from visuallm.elements.element_base import ElementBase
from visuallm.elements.plain_text_element import PlainTextElement


class InteractiveGenerationComponent(
    ComponentBase,
    DataPreparationMixin,
    ModelSelectionMixin,
    GenerationSelectorsMixin,
    MetricsMixin,
):
    def __init__(
        self,
        title: str = "Interactive Generation",
        model: Optional[PreTrainedModel] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        model_tokenizer_choices: Optional[MODEL_TOKENIZER_CHOICES] = None,
        metrics_on_generated_text: Dict[str, GENERATED_TEXT_METRIC] = {},
        metrics_on_probs: Dict[str, PROBS_METRIC] = {},
        dataset: Optional[DATASET_TYPE] = None,
        dataset_choices: Optional[DATASETS_TYPE] = None,
        selectors: SELECTORS_TYPE = {},
    ):
        self.main_heading_element = PlainTextElement(
            is_heading=True, heading_level=2, content=title
        )
        ModelSelectionMixin.__init__(
            self,
            on_model_change_callback=self.on_model_change_callback,
            model_tokenizer_choices=model_tokenizer_choices,
            model=model,
            tokenizer=tokenizer,
        )
        DataPreparationMixin.__init__(
            self,
            on_sample_change_callback=self.on_sample_change_callback,
            dataset=dataset,
            dataset_choices=dataset_choices,
        )
        GenerationSelectorsMixin.__init__(
            self,
            selectors=selectors,
            on_generation_changed_callback=self.on_sample_change_callback,
        )
        MetricsMixin.__init__(
            self,
            on_metrics_change=self.on_sample_change_callback,
            metrics_on_generated_text=metrics_on_generated_text,
            metrics_on_probs=metrics_on_probs,
        )
        input_display_elements = self.init_model_input_display()

        super().__init__(
            name="interactive_generation",
            title=title,
            elements=[
                self.main_heading_element,
                *self.dataset_elements,
                *self.model_elements,
                *self.generation_elements,
                *self.metrics_selection,
                *input_display_elements,
                *self.metrics_display,
            ],
        )

    def init_model_input_display(self) -> List[ElementBase]:
        self.input_display = PlainTextElement()
        return [self.input_display]

    def update_model_input_display(self):
        """Updated dataset sample is in `self._loaded_sample`"""
        self.input_display.content = self._loaded_sample

    def create_model_inputs(self):
        context = self.input_display.content
        model_inputs = self._tokenizer(context, return_tensors="pt")
        return model_inputs

    def create_target_encoding(self):
        context = self.input_display.content
        target = context + self.get_target_str()
        model_inputs = self._tokenizer(target, return_tensors="pt")
        return model_inputs.input_ids

    def get_target_str(self) -> str:
        return ""

    def update_generated_output_display(self):
        model_inputs = self.create_model_inputs()
        output = generate_output(
            model_inputs,
            tokenizer=self._tokenizer,
            model=self._model,
            **self.selected_generation_parameters,
        )
        decoded = decode_output(self._tokenizer, model_inputs, output)

        # compute mterics on generated
        probs, output_sequences = measure_output_probability(
            self._model, output.sequences, model_inputs
        )
        self.compute_metrics_on_predicted(
            decoded,
            [self.get_target_str() for _ in range(len(decoded))],
            probs,
            output_sequences,
        )

        # compute scores on target
        probs, output_sequences = measure_output_probability(
            self._model, self.create_target_encoding(), model_inputs
        )
        self.compute_metrics_on_target(self.get_target_str(), probs, output_sequences)

    def on_model_change_callback(self):
        self.data_force_update()
        self.dataset_callback()

    def on_sample_change_callback(self):
        self.update_model_input_display()
        self.update_generated_output_display()


def generate_output(
    model_inputs,
    *,
    tokenizer: PreTrainedTokenizer,
    model: PreTrainedModel,
    **kwargs,
):
    """Returns generated indices and log probabilities of generated indices."""

    # generate with loaded settings
    output = model.generate(
        **model_inputs,
        pad_token_id=tokenizer.eos_token_id,
        output_scores=True,
        return_dict_in_generate=True,
        **kwargs,
    )
    output = cast(GenerateOutput, output)

    return output


def decode_output(
    tokenizer: PreTrainedTokenizer,
    tokenized_inputs: BatchEncoding,
    output: GenerateOutput,
):
    detokenized_outputs = tokenizer.batch_decode(
        output.sequences, skip_special_tokens=True
    )
    detokenized_context = tokenizer.decode(
        tokenized_inputs.input_ids[0], skip_special_tokens=True
    )
    detokenized_outputs = [
        d.removeprefix(detokenized_context) for d in detokenized_outputs
    ]

    return detokenized_outputs


def measure_output_probability(
    model: PreTrainedModel,
    output_sequences: torch.Tensor,
    tokenized_inputs: BatchEncoding,
):
    """
    Returns:
        Tuple[List[float], List[float], List[float]]: log_probs,
            penalty_log_probs, probs
    """
    input_length = tokenized_inputs.input_ids.size(1)
    output_probs_list = []
    output_sequences_list = []
    for i in range(output_sequences.size(0)):
        with torch.no_grad():
            output = model(
                output_sequences[i : i + 1, :-1].contiguous(),
            )
            output_logits = output.logits
            output_probs = torch.softmax(cast(torch.Tensor, output_logits), dim=-1)
            output_probs = output_probs[:, input_length - 1 :, :]
            output_sequences_new = output_sequences[i : i + 1, input_length:]
        output_probs_list.append(output_probs)
        output_sequences_list.append(output_sequences_new)

    return output_probs_list, output_sequences_list



from pprint import pprint
from typing import Callable, List, MutableSet, Optional, Union

from flask import jsonify

from .elements.element_base import ElementBase, ElementWithEndpoint, URLNamedWrapper
from .elements.utils import register_named, sanitize_url


class ComponentBase:
    def __init__(
        self,
        name: str,
        title: str,
        elements: List[ElementBase],
        default_url: Optional[str] = None,
        default_callback: Optional[Callable] = None,
    ):
        if default_url is None:
            default_url = sanitize_url(name)
        self.default_url = default_url
        self.title = title
        self.name = name

        self.registered_element_names = set()
        self.registered_elements: List[ElementBase] = []

        self.registered_url_endpoints: MutableSet[str] = set()

        self.register_elements(elements)
        if default_callback is None:
            default_callback = self.fetch_info
        self.default_callback = default_callback

    def init_app(self, app):
        for element in self.registered_elements:
            # ensure that there are no two elements sharing the same url
            if isinstance(element, ElementWithEndpoint):
                register_named(element.get_url_named_wrapper(), app.registered_urls)
            element.add_endpoint(app)

        # ensure that there are no two components shared the same name
        register_named(self, app.registered_component_names)

        # ensure that there are no two components sharing the same default url
        register_named(URLNamedWrapper(self, "default_url"), app.registered_urls)
        app.add_endpoint(self.default_url, self.default_callback, methods=["GET"])

    def register_elements(
        self, elements: List[Union[ElementBase, ElementWithEndpoint]]
    ):
        for element in elements:
            # ensure that each element has a different name
            register_named(
                element, self.registered_element_names, self.registered_elements
            )

            if isinstance(element, ElementWithEndpoint):
                # ensure that there are no two url endpoints with the same name
                register_named(
                    element.get_url_named_wrapper(), self.registered_url_endpoints
                )
                element.parent_component = self

    def fetch_info(self, fetch_all: bool = True, debug_print: bool = False):
        res = dict(
            result="success",
            elementDescriptions=[
                element.construct_element_description()
                for element in self.registered_elements
                if element.changed or fetch_all
            ],
        )
        if debug_print:
            pprint(res)
        return jsonify(res)



from typing import Any, Callable, Dict, List, Tuple, Union

from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import (
    ButtonElement,
    CheckBoxSubElement,
    ChoicesSubElement,
    MinMaxSubElement,
    SelectorSubElement,
)

SELECTORS_TYPE = Dict[
    str,
    Union[
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float, float],
        Tuple[float, float, float, float],
        List[str],
        bool,
    ],
]


class GenerationSelectorsMixin:
    def __init__(
        self,
        selectors: SELECTORS_TYPE,
        on_generation_changed_callback: Callable[[], None],
    ):
        self.generation_selectors: List[SelectorSubElement] = []
        self.name_generation_selector_mapping: Dict[str, SelectorSubElement] = {}

        for selector_name, val in selectors.items():
            selector_text = str(selector_name)
            if not selector_text.rstrip().endswith(":"):
                selector_text = selector_text.rstrip() + ":"
            if isinstance(val, bool):
                selector = CheckBoxSubElement(selector_text, val)
            elif isinstance(val, tuple):
                if len(val) == 2:
                    selector = MinMaxSubElement(val[0], val[1], selector_text)
                elif len(val) == 3:
                    if isinstance(val[0], int):
                        selector = MinMaxSubElement(
                            val[0], val[1], selector_text, default_value=val[2]
                        )
                    else:
                        selector = MinMaxSubElement(
                            val[0], val[1], selector_text, step_size=val[2]
                        )
                elif len(val) == 4:
                    selector = MinMaxSubElement(
                        val[0],
                        val[1],
                        selector_text,
                        step_size=val[2],
                        default_value=val[3],
                    )
                else:
                    raise ValueError()
            elif isinstance(val, list):
                selector = ChoicesSubElement(val, selector_text)
            else:
                raise ValueError()
            self.generation_selectors.append(selector)
            self.name_generation_selector_mapping[selector_name] = selector

        self.generation_heading = PlainTextElement(
            content="Generation Parameters", is_heading=True
        )
        self._on_generation_changed_callback = on_generation_changed_callback
        self.generation_selector_button = ButtonElement(
            button_text="Send Generation Configuration",
            processing_callback=self._on_generation_changed_callback,
            subelements=self.generation_selectors,
        )

    @property
    def generation_elements(self):
        return [self.generation_heading, self.generation_selector_button]

    @property
    def selected_generation_parameters(self) -> Dict[str, Any]:
        return {
            name: value.selected
            for name, value in self.name_generation_selector_mapping.items()
        }

    def generation_callback(self):
        self._on_generation_changed_callback()
